// $Id: File.h,v 1.18 2003/11/25 23:13:27 vern Exp $
//
// Copyright (c) 1996, 1997, 1998, 1999, 2002
//      The Regents of the University of California.  All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that: (1) source code distributions
// retain the above copyright notice and this paragraph in its entirety, (2)
// distributions including binary code include the above copyright notice and
// this paragraph in its entirety in the documentation or other materials
// provided with the distribution, and (3) all advertising materials mentioning
// features or use of this software display the following acknowledgement:
// ``This product includes software developed by the University of California,
// Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
// the University nor the names of its contributors may be used to endorse
// or promote products derived from this software without specific prior
// written permission.
// THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
// WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

#ifndef file_h
#define file_h

#include <fcntl.h>
#include "util.h"
#include "Obj.h"
#include "Attr.h"

#ifdef USE_OPENSSL
# ifdef NEED_KRB5_H
#  include <krb5.h>
# endif // NEED_KRB5_H
# include "openssl/evp.h"
# include "openssl/pem.h"
# include "openssl/err.h"
#endif

class BroType;

class BroFile : public BroObj {
public:
	BroFile(FILE* arg_f);
	BroFile(FILE* arg_f, const char* filename, const char* access);
	BroFile(const char* filename, const char* access, BroType* arg_t = 0);
	virtual ~BroFile();

	// Returns false if an error occured.
	int Write(const char* data, int len = 0);

	void Flush()	{ fflush(f); }

	FILE* Seek(long position);	// seek to absolute position

	void SetBuf(bool buffered);	// false=line buffered, true=fully buffered

	const char* Name()	{ return name; }

	BroType* FType() const	{ return t; }

	int IsOpen() const	{ return is_open; }

	// Returns true if the close made sense, false if it was already
	// closed, not active, or whatever.
	int Close();

	void Describe(ODesc* d) const;

	// Rotate the logfile and runs post-processor if defined.
	void Rotate();

	// Set &rotate, &postprocessor attributes.
	void SetAttrs(Attributes* attrs);

	// Set rotate/postprocessor for all files that don't define them
	// by their own.
	static void SetDefaultRotation(double interval, const char* postproc);

	// Close all files which are managed by us.
	static void CloseManagedFiles();

	bool Serialize(Serializer* s) const;
	static BroFile* Unserialize(Serializer* s);

protected:
	void Init();

	BroFile* Prev()	{ return prev; }
	BroFile* Next()	{ return next; }
	void SetPrev(BroFile* f)	{ prev = f; }
	void SetNext(BroFile* f)	{ next = f; }

	void Suspend();
	void PurgeCache();
	void Unlink();
	void InsertAtBeginning();
	void MoveToBeginning();
	void InstallRotateTimer();

	// Returns nil if the file is not active, was in error, etc.
	// (Protected because we do not want anyone to write directly
	// to the file.)
	FILE* File();

	// Run post-processor with arguments
	// "<filename> <base-name> <time-of-open> <time-of-close>"
	void Postprocess(const char* file);

	// Initialize encryption with the given public key.
	void InitEncrypt(const char* keyfile);
	// Finalize encryption.
	void FinishEncrypt();

	FILE* f;
	BroType* t;
	char* name;
	char* access;
	int is_open;
	int is_managed;	// whether it's in the open-file cache
	long position;	// only valid if not is_open
	BroFile* next;	// doubly-linked list of open files
	BroFile* prev;
	Attributes* attrs;
	char* postproc;
	double rotate;
	Timer* rotate_timer;
	time_t open_time;
	bool dont_rotate; // See InstallRotateTimer()

#ifdef USE_OPENSSL
	EVP_PKEY* pub_key;
	EVP_CIPHER_CTX* cipher_ctx;

	static const int MIN_BUFFER_SIZE = 1024;
	unsigned char* cipher_buffer;
#endif

};

#endif
