// $Id: RemoteSerializer.cc,v 1.11 2003/11/19 07:26:01 vern Exp $
//
// Processes involved in the communication:
//
//	 (Local-Main) <-> (Local-Child) <-> (Remote-Child) <-> (Remote-Main)
//
// Message types:
//
// Communication protocol version
//	VERSION <version>
//
// Send serialization
//	SERIAL <ip> <port> <serialization>
//
// Terminate(d) connection
//	CLOSE  <ip> <port>
//
// Close(d) all connections
//	CLOSE_ALL
//
// Connect to remote side
//	CONNECTTO <ip> <port> <retry-interval> <use-ssl>
//
// Connected to remote side
//	CONNECTED <ip> <port>
//
// Request events from remote side
//	REQUEST <ip> <port> <list of events>
//
// Listen for non-SSL connections on ip/port (ip may be INADDR_ANY)
//	LISTEN_CLEAR  <ip> <port>
//
// Listen for SSL connection on ip/port (ip may be INADDR_ANY)
//	LISTEN_SSL  <ip> <port>
//
// Error caused by host
//	ERROR   <ip> <port> <msg>
//
// Some statistics about the given peer connection
//	STATS   <ip> <port> <string>
//
// Requests to set a new capture_filter
//	CAPTURE_FILTER <ip> <port> <string>
//
//
// Valid messages between processes:
//
//	Main -> Child
//		CONNECTTO
//		REQUEST
//		SERIAL
//		CLOSE
//		CLOSE_ALL
//		LISTEN_*
//		CAPTURE_FILTER
//	Child -> Main
//		CONNECTED
//		REQUEST
//		SERIAL
//		CLOSE
//		ERROR
//		STATS
//
//	Child <-> Child (<ip> and <port> fields are not set)
//		VERSION (transmits protocol version in port argument)
//		SERIAL
//		REQUEST
//		CAPTURE_FILTER

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <signal.h>
#include <strings.h>

#include "config.h"
#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <sys/resource.h>

#include "RemoteSerializer.h"
#include "Func.h"
#include "EventRegistry.h"
#include "Event.h"
#include "NetVar.h"
#include "Scope.h"

extern "C" {
#include "setsignal.h"
};

// Gets incremented each time there's an incompatible change
// to the communication internval.
static const unsigned short PROTOCOL_VERSION = 0x03;

static const char MSG_VERSION = 0x00;
static const char MSG_NONE = 0x01;
static const char MSG_SERIAL = 0x02;
static const char MSG_CLOSE = 0x03;
static const char MSG_CLOSE_ALL = 0x04;
static const char MSG_ERROR = 0x05;
static const char MSG_CONNECTTO = 0x06;
static const char MSG_CONNECTED = 0x07;
static const char MSG_REQUEST = 0x08;
static const char MSG_LISTEN_CLEAR = 0x09;
static const char MSG_LISTEN_SSL = 0x10;
static const char MSG_STATS = 0x11;
static const char MSG_CAPTURE_FILTER = 0x12;

// Start of every message between two processes.
typedef struct {
	char type;
	addr_type ip;
	uint16 port;
} CMsg;

static bool sendCMsg(ChunkedIO* io, char type, addr_type ip, uint16 port)
	{
	// We use the new[] operator here to avoid mismatches
	// when deleting the data.
	CMsg* msg = (CMsg*) new char[sizeof(CMsg)];

	msg->type = type;
	msg->ip = htonl(ip);
	msg->port = htons(port);

	ChunkedIO::Chunk* c = new ChunkedIO::Chunk;
	c->len = sizeof(CMsg);
	c->data = (char*) msg;

	return io->Write(c);
	}

static inline HashKey* buildHash(addr_type ip, uint16 port)
	{
	struct {
		unsigned long ip;

		// If we  use short here, gcc/Intel adds two
		// uninitialized alignment bytes. !@#$%
		unsigned long port;
	} tmp;

	tmp.ip = ip;
	tmp.port = port;

	return new HashKey(&tmp, sizeof(tmp));
	}

static inline const char* ip2a(uint32 ip)
	{
	static char buffer[32];
	struct in_addr addr;

	addr.s_addr = htonl(ip);

	return inet_ntop(AF_INET, &addr, buffer, 32);
	}

static pid_t child_pid = 0;

RemoteSerializer::RemoteSerializer()
	{
	current_ip = 0;
	current_port = 0;
	state = MSG_NONE;
	id_counter = 1;
	verbose = 0;
	listening = false;
	poll_time = 0;
	events_received = 0;
	events_sent = 0;

	error_fd = open("comm.log", O_WRONLY | O_CREAT | O_TRUNC, 0600);
	if ( error_fd < 0 )
		{
		error("Can't open comm.log: %s", strerror(errno));
		exit(1);
		}

	io = 0;
	}

RemoteSerializer::~RemoteSerializer()
	{
	if ( child_pid )
		{
		kill(child_pid, SIGQUIT);
		waitpid(child_pid, 0, 0);
		}

	delete io;
	}

void RemoteSerializer::Fork()
	{
	if ( child_pid )
		return;

	// If we are re-forking, remove old entries
	loop_over_list(peers, i)
		RemovePeer(peers[i]->ip, peers[i]->port);

	// Create pipe for communication between parent and child.
	int pipe[2];

	if ( socketpair(AF_UNIX, SOCK_STREAM, 0, pipe) < 0 )
		{
		Error(fmt("can't create pipe: %s", strerror(errno)));
		return;
		}

	child_pid = 0;

	int pid = fork();

	if ( pid < 0 )
		{
		Error(fmt("can't fork: %s", strerror(errno)));
		return;
		}

	if ( pid > 0 )
		{
		// Parent
		Log(fmt("Forked, parent pid is %d, child pid is %d", getpid(), pid));
		child_pid = pid;

		io = new ChunkedIOFd(pipe[0]);
		if ( ! io->Init() )
			{
			Error(fmt("can't init child io: %s", io->Error()));
			exit(1); // FIXME: Better way to handle this?
			}

		close(pipe[1]);

		return;
		}
	else
		{ // child
		SocketComm child(error_fd);

		ChunkedIOFd *io = new ChunkedIOFd(pipe[1]);
		if ( ! io->Init() )
			{
			Error(fmt("can't init parent io: %s", io->Error()));
			exit(1);
			}

		child.SetParentIO(io);
		close(pipe[0]);

		// Be nice.
		setpriority(PRIO_PROCESS, 0, 5);

		child.Run();
		internal_error("cannot be reached");
		}
	}

bool RemoteSerializer::Connect(addr_type ip, uint16 port, double r, bool use_ssl)
	{
	if ( ! child_pid )
		Fork();

	DEBUG_COMMPARENT("Sending MSG_CONNECTTO");
	if ( ! sendCMsg(io, MSG_CONNECTTO, ip, port) )
		{
		Error(fmt("can't send connect message: %s", io->Error()));
		return false;
		}

	// int's make more sense than double's here.
	unsigned int retry = (unsigned int) r;

	ChunkedIO::Chunk* c = new ChunkedIO::Chunk;
	c->len = sizeof(retry) + sizeof(use_ssl);
	c->data = new char[c->len];
	*(unsigned int*)c->data = (unsigned int) retry;
	*(bool*)(c->data + sizeof(unsigned int)) = (unsigned int) use_ssl;

	if ( ! io->Write(c) )
		{
		Error(fmt("can't send connect data: %s", io->Error()));
		return false;
		}

	AddPeer(ip, port, false);

	return true;
	}

bool RemoteSerializer::Request(addr_type ip, uint16 port, RE_Matcher* pattern)
	{
	EventRegistry::string_list* handlers = event_registry->Match(pattern);
	ChunkedIO::Chunk* c = new ChunkedIO::Chunk;

	// Concat all handler names and put them into a Chunk.
	c->len = 0;
	loop_over_list(*handlers, i)
		c->len += strlen((*handlers)[i]) + 1;

	if ( ! c->len )
		{
		Log("Warning: No events to request");
		return true;
		}

	c->data = new char[c->len];
	char* d = c->data;
	loop_over_list(*handlers, j)
		{
		for ( const char* p = (*handlers)[j]; *p; *d++ = *p++ )
			;
		*d++ = '\0';
		}

	delete handlers;

	DEBUG_COMMPARENT("Sending MSG_REQUEST");
	if ( ! sendCMsg(io, MSG_REQUEST, ip, port) )
		{
		Error(fmt("can't send request message: %s", io->Error()));
		return false;
		}

	if ( ! io->Write(c) )
		{
		Error(fmt("can't send request data: %s", io->Error()));
		return false;
		}

	return true;
	}

bool RemoteSerializer::SendEvent(addr_type ip, uint16 port,
					const char* name, val_list* vl)
	{
	Peer* p = LookupPeer(ip, port, false);
	if ( ! p )
		{
		warn("ignoring request to send ID to unknown client");
		return true;
		}

	DEBUG_COMMPARENT("Sending MSG_SERIAL");
	if ( ! sendCMsg(io, MSG_SERIAL, ip, port) )
		{
		Error(fmt("can't send serialization message: %s", io->Error()));
		return false;
		}

	++events_sent;

	SetCache(&p->cache_out);
	return Serialize(name, vl);
	}

bool RemoteSerializer::SendID(addr_type ip, uint16 port, const char* name)
	{
	ID* id = global_scope()->Lookup(name);
	if ( ! name )
		{
		warn(fmt("ignoring request to send unknown ID %s", name));
		return true;
		}

	Peer* p = LookupPeer(ip, port, false);
	if ( ! p )
		{
		warn("ignoring request to send ID to unknown client");
		return true;
		}

	DEBUG_COMMPARENT("Sending MSG_SERIAL");
	if ( ! sendCMsg(io, MSG_SERIAL, ip, port) )
		{
		Error(fmt("can't send serialization message: %s", io->Error()));
		return false;
		}

	SetCache(&p->cache_out);
	SerialInfo info;
	return Serialize(id, info);
	}

bool RemoteSerializer::SendCaptureFilter(addr_type ip, uint16 port,
						const char* filter)
	{
	DEBUG_COMMPARENT("Sending MSG_CAPTURE_FILTER");
	if ( ! sendCMsg(io, MSG_CAPTURE_FILTER, ip, port) )
		{
		Error(fmt("can't send capture_filter message: %s", io->Error()));
		return false;
		}

	ChunkedIO::Chunk* c = new ChunkedIO::Chunk;
	c->len = strlen(filter) + 1;
	c->data = new char[c->len];
	strcpy(c->data, filter);

	if ( ! io->Write(c) )
		{
		Error(fmt("can't send capture_filter data: %s", io->Error()));
		return false;
		}

	return true;
	}

bool RemoteSerializer::Listen(addr_type ip, uint16 port, bool expect_ssl)
	{
#ifndef USE_OPENSSL
	if ( expect_ssl )
		{
		Error("listening for SSL connections requested, but SSL support is not compiled in");
		return false;
		}
#endif

	if ( ! child_pid )
		Fork();

	DEBUG_COMMPARENT("Sending MSG_LISTEN*");
	if ( ! sendCMsg(io, expect_ssl ? MSG_LISTEN_SSL : MSG_LISTEN_CLEAR, ip, port) )
		{
		Error(fmt("can't send listen message: %s", io->Error()));
		return false;
		}

	listening = true;
	return true;
	}

bool RemoteSerializer::Poll(bool may_block)
	{
	if ( ! child_pid )
		return true;

	switch ( state ) {
	case MSG_NONE:
		{
		// Polling needs CPU time, so reduce it to a
		// reasonable amount.
		if ( network_time - poll_time < 0.1 &&
		     network_time != poll_time )
			return true;

		poll_time = network_time;

		// CMsg follows:
		ChunkedIO::Chunk* c;
		if ( ! io->Read(&c, may_block) )
			{
			Error(fmt("can't read child's cmsg: %s",
					io->Error()));
			return false;
			}

		if ( ! c )
			return true;

		CMsg* msg = (CMsg*) c->data;
		current_ip = ntohl(msg->ip);
		current_port = ntohs(msg->port);

		switch ( msg->type ) {
		case MSG_CONNECTED:
			DEBUG_COMMPARENT("Got MSG_CONNECTED");
			if ( ! LookupPeer(current_ip, current_port, true) )
				{
				AddPeer(current_ip, current_port, true);
				RaiseEvent(remote_connection_established,
					current_ip, current_port);
				}

			SendID(current_ip, current_port, "event_source_description");
			break;

		case MSG_CLOSE:
			DEBUG_COMMPARENT("Got MSG_CLOSE");
			if ( LookupPeer(current_ip, current_port, true) )
				{
				RaiseEvent(remote_connection_closed, current_ip, current_port);
				// Change state to disconneced.
				AddPeer(current_ip, current_port, false);
				}
			break;

		default:
			DEBUG_COMMPARENT("Got MSG_*");
			// Msg with more arguments
			state = msg->type;
		}

		delete [] c->data;
		delete c;

		return true;
		}

	case MSG_SERIAL:
		// Serialization follows
		return ProcessSerialization();

	case MSG_REQUEST:
		// List of events follows
		return ProcessRequestMsg();

	case MSG_ERROR:
		return ProcessErrorMsg();

	case MSG_STATS:
		return ProcessStatsMsg();

	case MSG_CAPTURE_FILTER:
		return ProcessCaptureFilterMsg();

	default:
		internal_error(fmt("Unexpected msg type: %d", (int) state));
	}
	}

RemoteSerializer::Peer* RemoteSerializer::AddPeer(addr_type ip, uint16 port,
					bool connected)
	{
	HashKey* key = buildHash(ip, port);
	Peer* p = peers_byaddr.Lookup(key);
	if ( p )
		{
		delete key;
		if ( p->connected != connected )
			{
			p->connected = connected;
			Log(fmt("Peer #%d now %sconnected",
				p->id, (connected ? "" : "un")));

			if ( ! connected )
				UnregisterHandlers(p);
			else
				{
				p->cache_in.Clear();
				p->cache_out.Clear();
				}
			}

		return p;
		}

	// Make empty list of registered handlers.
	p = new Peer;
	p->id = id_counter++;
	p->ip = current_ip = ip;
	p->port = current_port = port;
	p->connected = connected;
	p->val = new RecordVal(event_source);
	p->val->Assign(0, new Val(p->id, TYPE_COUNT));
	p->val->Assign(1, new AddrVal(htonl(p->ip)));
	p->val->Assign(2, new PortVal(p->port));
	p->val->Assign(3, new StringVal(""));	// set when received

	peers_byaddr.Insert(key, p);
	peers.append(p);
	delete key;

	Log(fmt("Added peer #%d (%sconnected)", p->id, (connected ? "" : "un" )));

	return p;
	}

void RemoteSerializer::UnregisterHandlers(Peer *peer)
	{
	// Unregister the peers for the EventHandlers.
	loop_over_list(peer->handlers, i)
		{
		peer->handlers[i]->RemoveRemoteHandler(peer->ip, peer->port);
		}
	}

Event::SourceID RemoteSerializer::RemovePeer(addr_type ip, uint16 port)
	{
	HashKey* key = buildHash(ip, port);
	Peer* p = peers_byaddr.Lookup(key);

	if ( ! p )
		{
		Log("Warning: removing peer which doesn't exist");
		return 0;
		}

	peers_byaddr.Remove(key);
	peers.remove(p);
	delete key;

	UnregisterHandlers(p);

	current_ip = p->ip;
	current_port = p->port;
	Log("Removed peer");

	int id = p->id;
	Unref(p->val);
	delete p;

	return id;
	}

RemoteSerializer::Peer* RemoteSerializer::LookupPeer(addr_type ip, uint16 port, bool only_if_connected)
	{
	HashKey* key = buildHash(ip, port);
	Peer* p = peers_byaddr.Lookup(key);
	delete key;

	if ( p && (p->connected || ! only_if_connected) )
		return p;
	else
		return 0;
	}

bool RemoteSerializer::ProcessRequestMsg()
	{
	DEBUG_COMMPARENT("Processing request");

	ChunkedIO::Chunk* c;
	if ( ! io->Read(&c) )
		{
		Error(fmt("can't read event requests: %s", io->Error()));
		return false;
		}

	if ( ! c )
		return true;

	Peer* peer = LookupPeer(current_ip, current_port, true);
	if ( ! peer )
		return false;

	// Unregister old handlers.
	UnregisterHandlers(peer);

	// Register new handlers.
	char* p = c->data;
	while ( p < c->data + c->len )
		{
		EventHandler* handler = event_registry->Lookup(p);
		if ( handler )
			{
			handler->AddRemoteHandler(current_ip, current_port);
			peer->handlers.append(handler);
			Log(fmt("Registered for event %s", p));
			}
		else
			Log(fmt("Request for unknown event %s", p));

		p += strlen(p) + 1;
		}

	delete [] c->data;
	delete c;

	current_ip = 0;
	current_port = 0;
	state = MSG_NONE;
	return true;
	}

bool RemoteSerializer::ProcessErrorMsg()
	{
	DEBUG_COMMPARENT("Processing error");

	ChunkedIO::Chunk* c;
	if ( ! io->Read(&c) )
		{
		Error(fmt("can't read error msg: %s", io->Error()));
		return false;
		}

	if ( ! c )
		return true;

	ReportError(c->data);

	delete [] c->data;
	delete c;

	current_ip = 0;
	current_port = 0;
	state = MSG_NONE;
	return true;
	}

bool RemoteSerializer::ProcessStatsMsg()
	{
	DEBUG_COMMPARENT("Processing stats");

	ChunkedIO::Chunk* c;
	if ( ! io->Read(&c) )
		{
		Error(fmt("can't read stats msg: %s", io->Error()));
		return false;
		}

	if ( ! c )
		return true;

	// Take the opportunity to log our stats, too.
	LogStats();

	// Split the concatenated child stats into indiviual log messages.
	int count = 0;
	for( char* p = c->data; p < c->data + c->len; p += strlen(p) + 1 )
		Log(fmt("Child : [%d] %s", count++, p));

	delete [] c->data;
	delete c;

	current_ip = 0;
	current_port = 0;
	state = MSG_NONE;

	return true;
	}

bool RemoteSerializer::ProcessCaptureFilterMsg()
	{
	ChunkedIO::Chunk* c;
	if ( ! io->Read(&c) )
		{
		Error(fmt("can't read capture filter msg: %s", io->Error()));
		return false;
		}

	if ( ! c )
		return true;

	RaiseEvent(remote_capture_filter, current_ip, current_port, c->data);

	delete [] c->data;
	delete c;

	current_ip = 0;
	current_port = 0;
	state = MSG_NONE;

	return true;
	}

bool RemoteSerializer::ProcessSerialization()
	{
	DEBUG_COMMPARENT("Processing serialization");

	Peer* p = LookupPeer(current_ip, current_port, false);
	if ( ! p )
		{
		Error("serialization from unknown peer");
		return false;
		}

	SetCache(&p->cache_in);
	SerialInfo info;
	int i = Unserialize(&info);
	if ( i > 0 )
		state = MSG_NONE;

	return i >= 0;
	}

void RemoteSerializer::GotEvent(const char* event, double time,
				EventHandlerPtr func, val_list* args)
	{
	DEBUG_COMMPARENT("Got event");

	Peer* peer = LookupPeer(current_ip, current_port, true);
	if ( ! peer )
		{
		Error("unserialized event from unknown source");
		return;
		}

	BufferedEvent* e = new BufferedEvent;

	e->time = ::time(0);	// our time, not the time when event generated
	e->src = peer->id;
	e->handler = func;
	e->args = args;

	events.append(e);
	++events_received;
	}

void RemoteSerializer::GotID(ID* id, Val* val)
	{
	if ( strcmp(id->Name(), "event_source_description" ) == 0 )
		 {
		 Peer* peer = LookupPeer(current_ip, current_port, true);
		 if ( ! peer )
			 {
			 Error("unserialized id from unknown source");
			 return;
			 }

		 peer->val->Assign(3, val);
		 return;
		 }

	run_time("received unexpected ID over the network");
	}

void RemoteSerializer::UnknownID(ID* id)
	{
	run_time("received unknown ID over the network");
	}

void RemoteSerializer::DispatchNextEvent()
	{
	if ( events.length() )
		{
		BufferedEvent* be = events[0];
		::Event* event = new ::Event(be->handler, be->args, be->src);

		mgr.Dispatch(event);
		delete be;
		events.remove_nth(0);
		}
	}

double RemoteSerializer::NextTimestamp()
	{
	return events.length() ? events[0]->time : 0;
	}

bool RemoteSerializer::IsActive()
	{
	return listening || peers.length();
	}

void RemoteSerializer::ReportError(const char* msg)
	{
	RaiseEvent(remote_connection_error, current_ip, current_port, msg);
	Log(msg);
	}

void RemoteSerializer::Log(const char* msg)
	{
	const int BUFSIZE = 1024;
	char buffer[BUFSIZE];

	int len = snprintf(buffer, BUFSIZE, "%lu (parent) ", time(0));

	if ( current_ip )
		len += snprintf(buffer + len, BUFSIZE - len,
				"[%s:%d] ", ip2a(current_ip), current_port);

	len += snprintf(buffer + len, BUFSIZE - len, "%s\n ", msg);

	write(error_fd, buffer, len);
	}

void RemoteSerializer::RaiseEvent(EventHandlerPtr event, addr_type ip,
				uint16 port, const char* arg)
	{
	val_list* vl = new val_list;

	// Sic! Network order for AddrVal, host order for PortVal.
	vl->append(new AddrVal(htonl(ip)));
	vl->append(new PortVal(port));
	if ( arg )
		vl->append(new StringVal(arg));

	// If we only have remote sources, the network time
	// will not increase as long as no peers are connected.
	// Therefore, we send these events immediately.
	mgr.Dispatch(new Event(event, vl));
	}

void RemoteSerializer::LogStats()
	{
	char buffer[512];
	assert(io);
	io->Stats(buffer, 512);
	Log(fmt("Parent: %s events=%lu/%lu",
		buffer, events_received, events_sent));
	}

RecordVal *RemoteSerializer::GetEventSourceVal(Event::SourceID id)
	{
	// Is it worh it to use a map here indexed by ID rather than a list?
	// Probably not as long as we have only a couple of peers. Same
	// argumentation would work for the peer_map, though...
	loop_over_list(peers, i)
		if ( peers[i]->id == id )
			{
			Ref(peers[i]->val);
			return peers[i]->val;
			}

	return 0;
	}


////////////////////////////

// If true (set by signal handler), we will log some stats to parent.
static bool log_stats = false;

// How often stats are sent (in seconds).
// Perhaps we should make this configurable...
const int STATS_INTERVAL = 60;

static RETSIGTYPE sig_handler(int signo)
	{
	// SIGALRM is the only one we get.
	log_stats = true;
	}

SocketComm::SocketComm(int arg_error_fd)
	{
	io = 0;
	error_fd = arg_error_fd;
	peer = 0;
	parent_state = MSG_NONE;
	listen_fd_clear = -1;
	listen_fd_ssl = -1;

	// We don't want to use the signal handlers of our parent.
	(void) setsignal(SIGTERM, SIG_DFL);
	(void) setsignal(SIGINT, SIG_DFL);
	(void) setsignal(SIGUSR1, SIG_DFL);
	(void) setsignal(SIGUSR2, SIG_DFL);
	(void) setsignal(SIGCONT, SIG_DFL);
	(void) setsignal(SIGCHLD, SIG_DFL);

	(void) setsignal(SIGALRM, sig_handler);
	alarm(STATS_INTERVAL);
	}

SocketComm::~SocketComm()
	{
	loop_over_list(peers, i)
		delete peers[i]->io;

	delete io;
	close(error_fd);
	close(listen_fd_clear);
	close(listen_fd_ssl);
	}

void SocketComm::Run()
	{
	while ( true )
		{
		// Logging signaled?
		if ( log_stats )
			{
			LogStats();
			log_stats = false;
			alarm(STATS_INTERVAL);
			}

		// Build FDSets for select.
		fd_set fd_read, fd_write, fd_except;

		FD_ZERO(&fd_read);
		FD_ZERO(&fd_write);
		FD_ZERO(&fd_except);

		int max_fd = 0;

		FD_SET(io->Fd(), &fd_read);
		max_fd = io->Fd();

		loop_over_list(peers, i)
			{
			if ( peers[i]->connected )
				{
				FD_SET(peers[i]->io->Fd(), &fd_read);
				if ( peers[i]->io->Fd() > max_fd )
					max_fd = peers[i]->io->Fd();
				}
			else
				{
				if ( peers[i]->next_try > 0 && time(0) > peers[i]->next_try )
					// Try reconnect.
					Connect(peers[i]->ip, peers[i]->port, peers[i]->retry, peers[i]->ssl);
				}
			}

		if ( listen_fd_clear >= 0 )
			{
			FD_SET(listen_fd_clear, &fd_read);
			if ( listen_fd_clear > max_fd )
				max_fd = listen_fd_clear;
			}

		if ( listen_fd_ssl >= 0 )
			{
			FD_SET(listen_fd_ssl, &fd_read);
			if ( listen_fd_ssl > max_fd )
				max_fd = listen_fd_ssl;
			}

		// We cannot rely solely on select() as the there may
		// be some data left in our input queues. So, we use
		// a small timeout for select and check for data
		// manually afterwards.
		struct timeval small_timeout;
		small_timeout.tv_sec = 0;
		small_timeout.tv_usec = 10;

		if ( select(max_fd + 1, &fd_read, &fd_write, &fd_except,
				&small_timeout) < 0 )
			// Ignore errors for now.
			continue;

		while ( io->CanRead() )
			ProcessParentMessage();

		loop_over_list(peers, j)
			{
			// We have to be careful here as the peer may
			// be removed when an error occurs.
			Peer* current = peers[j];
			while ( j < peers.length() && peers[j] == current &&
				current->connected && current->io->CanRead() )
				{
				ProcessRemoteMessage(current);
				}
			}

		if ( listen_fd_clear >= 0 && FD_ISSET(listen_fd_clear, &fd_read) )
			AcceptConnection(listen_fd_clear);

		if ( listen_fd_ssl >= 0 && FD_ISSET(listen_fd_ssl, &fd_read) )
			AcceptConnection(listen_fd_ssl);

		}
	}


void SocketComm::ProcessParentMessage()
	{
	DEBUG_COMMCHILD("Processing parent message");

	switch ( parent_state ) {
	case MSG_NONE:
		{
		// CMsg follows.
		ChunkedIO::Chunk* c;
		if ( ! io->Read(&c) )
			{
			if ( io->Eof() )
				{
				Log("Parent died");
				Kill();
				}

			Error(fmt("can't read parent's cmsg: %s",
					io->Error()), 0, true);
			return;
			}

		if ( ! c )
			return;

		CMsg* msg = (CMsg*) c->data;

		ip = ntohl(msg->ip);
		port = ntohs(msg->port);

		HashKey* key = buildHash(ip, port);
		peer = peers_byaddr.Lookup(key);

		delete key;

		switch ( msg->type ) {
		case MSG_LISTEN_CLEAR:
			DEBUG_COMMCHILD("Got MSG_LISTEN_CLEAR from parent");
			Listen(ip, port, false);
			break;

		case MSG_LISTEN_SSL:
			DEBUG_COMMCHILD("Got MSG_LISTEN_SSL from parent");
			Listen(ip, port, true);
			break;

		case MSG_CLOSE:
			{
			DEBUG_COMMCHILD("Got MSG_CLOSE from parent");
			if ( peer && peer->connected )
				CloseConnection(peer, false);
			break;
			}

		case MSG_CLOSE_ALL:
			{
			DEBUG_COMMCHILD("Got MSG_CLOSEALL from parent");
			loop_over_list(peers, i)
				{
				if ( peers[i]->connected )
					CloseConnection(peers[i], false);
				}
			break;
			}

		default:
			DEBUG_COMMCHILD("Got MSG_* from parent");
			// Msg with more arguments.
			parent_state = msg->type;
		}

		delete [] c->data;
		delete c;
		break;
		}

	case MSG_CONNECTTO:
		{
		DEBUG_COMMCHILD("Processing MSG_CONNNECTTO arguments");

		ChunkedIO::Chunk* c;
		if ( ! io->Read(&c) )
			return;

		if ( ! c )
			return;

		unsigned int* retry = (unsigned int*) c->data;
		bool* use_ssl = (bool *)(c->data + sizeof(unsigned int));

		if ( ! peer )
			Connect(ip, port, *retry, *use_ssl);

		delete [] c->data;
		delete c;

		parent_state = MSG_NONE;
		break;
		}

	case MSG_REQUEST:
	case MSG_SERIAL:
	case MSG_CAPTURE_FILTER:
		{ // we simply forward it to the remote side
		DEBUG_COMMCHILD("Processing MSG_REQUEST/SERIAL arguments");

		ChunkedIO::Chunk* c;
		if ( ! io->Read(&c) || ! c )
			return;

		if ( peer && peer->connected )
			{
			DEBUG_COMMCHILD("Forwarding to remote side");

			if ( ! sendCMsg(peer->io, parent_state, 0, 0) )
				{
				Error(fmt("can't send message to remote: %s",
						peer->io->Error()), 0, true);
				return;
				}

			if ( ! peer->io->Write(c) )
				{
				Error(fmt("can't forward chunk to remote: %s",
						peer->io->Error()), 0, true);
				return;
				}
			}

		else
			Error("not connected to destination of message",
				0, false);

		parent_state = MSG_NONE;
		break;
		}

	default:
		internal_error("unexpected state");
	}
	}

void SocketComm::ProcessRemoteMessage(Peer* p)
	{
	DEBUG_COMMCHILD("Processing remote message");

	peer = p;

	switch ( p->state ) {
	case MSG_NONE:
		{ // CMsg follows
		ChunkedIO::Chunk* c;
		if ( ! p->io->Read(&c) )
			{
			if ( p->io->Eof() )
				{
				Log("Peer closed connection");
				CloseConnection(p, true);
				}
			else
				Error(fmt("can't read peer's cmsg: %s",
					p->io->Error()), p, false);
			return;
			}

		if ( ! c )
			return;

		CMsg* msg = (CMsg*) c->data;

		if ( msg->type == MSG_VERSION )
			{
			DEBUG_COMMCHILD("Got MSG_VERSION from remote");
			if ( msg->port != htons(PROTOCOL_VERSION) )
				Error(fmt("protocol version mismatch: Got %d, but expected %d\n",
						msg->port, PROTOCOL_VERSION),
					peer, false);
			break;
			}

		DEBUG_COMMCHILD("Got MSG_* from remote");

		peer->state = msg->type;

		delete[] c->data;
		delete c;

		break;
		}

	case MSG_REQUEST:
	case MSG_SERIAL:
	case MSG_CAPTURE_FILTER:
		{
		DEBUG_COMMCHILD("Forwarding to parent");

		// We simply forward it to our parent
		ChunkedIO::Chunk* c;
		if ( ! peer->io->Read(&c) )
			{
			Error(fmt("can't read chunk from peer: %s",
				io->Error()), peer, false);
			return;
			}

		if ( ! c )
			return;

		if ( ! sendCMsg(io, peer->state, peer->ip, peer->port) )
			{
			Error(fmt("can't send message to parent: %s",
					io->Error()), NULL, true);
			return;
			}

		if ( ! io->Write(c) )
			{
			Error(fmt("can't forward chunk to parent: %s",
					io->Error()), NULL, true);
			return;
			}

		peer->state = MSG_NONE;
		break;
		}

	default:
		internal_error("Unexpected state");
	}
	}

SocketComm::Peer* SocketComm::Connect(uint32 dst, uint16 port,
					unsigned int retry, bool use_ssl)
	{
	struct sockaddr_in server;

	int sockfd = socket(PF_INET, SOCK_STREAM, 0);
	if ( sockfd < 0 )
		{
		Error(fmt("can't create socket, %s", strerror(errno)), 0, false);
		return 0;
		}

	bzero(&server, sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port = htons(port);
	server.sin_addr.s_addr = htonl(dst);

	bool connected;

	if ( connect(sockfd, (sockaddr*) &server, sizeof(server)) < 0 )
		{
		connected = false;
		Log(fmt("Connect failed, %s", strerror(errno)));
		}
	else
		connected = true;

	if ( ! (connected || retry) )
		{
		DEBUG_COMMCHILD("Sending MSG_CLOSE");

		if ( ! sendCMsg(io, MSG_CLOSE, dst, port) )
			Error(fmt("can't send close message: %s",
					io->Error()), 0, true);
		return 0;
		}

	HashKey* hash = buildHash(dst, port);
	peer = peers_byaddr.Lookup(hash);
	if ( ! peer )
		{
		peer = new Peer;
		peer->ip = dst;
		peer->port = port;
		peer->retry = retry;
		peer->ssl = use_ssl;

		peers.append(peer);
		peers_byaddr.Insert(hash, peer);
		}

	peer->connected = connected;
	peer->next_try = connected ? 0 : time(0) + retry;
	peer->state = MSG_NONE;
	peer->io = 0;

	if ( connected )
		{
		if ( use_ssl )
			{
#ifdef USE_OPENSSL
			peer->io = new ChunkedIOSSL(sockfd, false);
#else
			run_time("SSL connection requested, but SSL support not compiled in");
			delete peer;
			delete hash;
			return 0;
#endif
			}
		else
			peer->io = new ChunkedIOFd(sockfd);

		if ( ! peer->io->Init() )
			{
			Error(fmt("can't init peer io: %s",
					peer->io->Error()), peer, false);
			delete hash;
			return 0;
			}
		}

	delete hash;

	if ( connected )
		{
		Log("Connected");
		DEBUG_COMMCHILD("Sending MSG_CONNECTED");

		if ( ! sendCMsg(io, MSG_CONNECTED, peer->ip, peer->port) )
			Error(fmt("can't send MSG_CONNECTED: %s",
					io->Error()), peer, false);

		DEBUG_COMMCHILD("Sending MSG_VERSION");
		if ( ! sendCMsg(peer->io, MSG_VERSION, 0, PROTOCOL_VERSION) )
			Error(fmt("can't send MSG_VERSION: %s",
					  peer->io->Error()), peer, false);
		}

	return connected ? peer : 0;
	}

void SocketComm::CloseConnection(Peer* p, bool reconnect)
	{
	if ( ! p->connected )
		return;

	Log("Closed connection");

	DEBUG_COMMCHILD("Sending MSG_CLOSE");

	if ( ! sendCMsg(io, MSG_CLOSE, p->ip, p->port) )
		Error("can't send MSG_CLOSE", 0, false);

	if ( ! ( p->retry && reconnect ) )
		{
		HashKey* key = buildHash(p->ip, p->port);
		peers_byaddr.Remove(key);
		delete key;

		peers.remove(p);
		delete p->io;
		delete p;
		}
	else
		{
		delete p->io;
		p->io = 0;
		p->connected = false;
		p->next_try = time(0) + p->retry;
		}

	peer = 0;
	}

void SocketComm::Listen(addr_type ip, uint16 port, bool expect_ssl)
	{
	int* listen_fd = expect_ssl ? &listen_fd_ssl : &listen_fd_clear;

	if ( *listen_fd >= 0 )
		close(*listen_fd);

	struct sockaddr_in server;

	*listen_fd = socket(PF_INET, SOCK_STREAM, 0);
	if ( *listen_fd < 0 )
		{
		Error(fmt("can't create listen socket, %s",
				strerror(errno)), 0, false);
		return;
		}

	// Set SO_REUSEADDR.
	int turnon = 1;
	if ( setsockopt(*listen_fd, SOL_SOCKET, SO_REUSEADDR,
			&turnon, sizeof(turnon)) < 0 )
		{
		Error(fmt("can't set SO_REUSEADDR, %s",
				strerror(errno)), 0, false);
		return;
		}

	bzero(&server, sizeof(server));
	server.sin_family = AF_INET;
	server.sin_port = htons(port);
	server.sin_addr.s_addr = htonl(ip);

	if ( bind(*listen_fd, (sockaddr*) &server, sizeof(server)) < 0 )
		{
		Error(fmt("can't bind to port, %s", strerror(errno)), 0, false);
		*listen_fd = -1;
		return;
		}

	if ( listen(*listen_fd, 50) < 0 )
		{
		Error(fmt("can't listen, %s", strerror(errno)), 0, false);
		return;
		}

	Log(fmt("Listening on %s:%d", ip2a(ip), port));
	}

void SocketComm::AcceptConnection(int fd)
	{
	sockaddr_in client;
	socklen_t len = sizeof(client);

	int clientfd = accept(fd, (sockaddr*) &client, &len);
	if ( clientfd >= 0 )
		{
		peer = new Peer;
		peer->ip = ntohl(client.sin_addr.s_addr);
		peer->port = ntohs(client.sin_port);
		peer->state = MSG_NONE;
		peer->connected = true;
		peer->retry = 0;
		peer->ssl = (fd == listen_fd_ssl);
		peer->next_try = 0;

#ifdef USE_OPENSSL
		if ( peer->ssl )
			peer->io = new ChunkedIOSSL(clientfd, true);
		else
			peer->io = new ChunkedIOFd(clientfd);
#else
		assert( ! peer->ssl );
		peer->io = new ChunkedIOFd(clientfd);
#endif

		if ( ! peer->io->Init() )
			{
			Error(fmt("can't init peer io: %s",
					peer->io->Error()), peer, false);
			return;
			}

		peers.append(peer);
		HashKey* hash = buildHash(peer->ip, peer->port);
		peers_byaddr.Insert(hash, peer);
		}
	else
		{
		Error(fmt("accept failed, %s %d",
				strerror(errno), errno), 0, false);
		return;
		}

	Log(fmt("Accepted %s connection", (peer->ssl ? "SSL" : "clear")));

	DEBUG_COMMCHILD("Sending MSG_CONNECTED");
	if ( ! sendCMsg(io, MSG_CONNECTED, peer->ip, peer->port) )
		Error("can't send MSG_CONNECTED", peer, false);

	DEBUG_COMMCHILD("Sending MSG_VERSION");
	if ( ! sendCMsg(peer->io, MSG_VERSION, 0, PROTOCOL_VERSION) )
		Error(fmt("can't send MSG_VERSION: %s",
					peer->io->Error()), peer, false);
	}

void SocketComm::Error(const char* msg, Peer* p, bool kill_me)
	{
	Log(msg);

	// If a remote peer causes an error, we shutdown the connection
	// as resynchronizing is in general not possible. But we may
	// try again later.
	if ( peer && peer->connected )
		CloseConnection(peer, true);

	if ( kill_me )
		Kill();
	}

void SocketComm::Log(const char* msg)
	{
	const char* source = "<local>";
	if ( peer )
		source = ip2a(peer->ip);

	char buffer[1024];
	int len = snprintf(buffer, sizeof(buffer), "%lu (child) %s: %s\n",
				time(0), source, msg);
	write(error_fd, buffer, len);
	}

void SocketComm::Kill()
	{
	Log("terminating");

	close(error_fd);
	close(listen_fd_clear);
	close(listen_fd_ssl);

	kill(getpid(), SIGQUIT);

	while ( 1 )
		; // loop until killed
	}

void SocketComm::LogStats()
	{
	if ( ! peers.length() )
		return;

	// Concat stats of all peers into single buffer.
	char* buffer = new char[peers.length() * 512];
	int pos = 0;

	loop_over_list(peers, i)
		{
		if ( peers[i]->connected )
			peers[i]->io->Stats(buffer+pos, 512);
		else
			strcpy(buffer+pos, "not connected");
		pos += strlen(buffer+pos) + 1;
		}

	// Send it.
	if ( ! sendCMsg(io, MSG_STATS, peer->ip, peer->port) )
		Error("can't send MSG_STATS", 0, false);

	ChunkedIO::Chunk* c = new ChunkedIO::Chunk;
	c->data = buffer;
	c->len = pos;

	if ( ! io->Write(c) )
		Error(fmt("can't send stat string: %s", strerror(errno)), 0, true);
	}
