// $Id: Serializer.cc,v 1.11 2003/12/03 00:34:28 vern Exp $

#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <dirent.h>
#include <libgen.h>
#include <sys/time.h>
#include <sys/stat.h>

#include "Serializer.h"
#include "Scope.h"
#include "Stmt.h"
#include "Logger.h"
#include "Func.h"
#include "Event.h"
#include "EventRegistry.h"
#include "SerializationFormat.h"
#include "NetVar.h"

Serializer::Serializer(SerializationFormat* arg_format)
	{
	if ( arg_format )
		format = arg_format;
	else
		format = new BinarySerializationFormat();

	io = 0;
	verbose = false;
	error_descr = 0;
	current_cache = 0;
	print_only = false;
	}

Serializer::~Serializer()
	{
	delete format;
	delete [] error_descr;
	}

bool Serializer::Read(string* s)
	{
	const char* cstr;
	int len;
	if ( format->Read(&cstr, &len) )
		{
		s->assign(cstr, len);
		delete [] cstr;
		return true;
		}
	else
		return false;
	}

void Serializer::StartSerialization()
	{
	format->StartWrite();
	}

bool Serializer::EndSerialization()
	{
	ChunkedIO::Chunk* chunk = new ChunkedIO::Chunk;
	chunk->len = format->EndWrite(&chunk->data);
	return io->Write(chunk);
	}

bool Serializer::Serialize(ID* id, const SerialInfo& info)
	{
	assert(current_cache);

	SetErrorDescr(fmt("serializing ID %s", id->Name()));

	if ( ! (id->IsGlobal() || id->IsEnumConst()) )
		{
		Error("non-global identifiers cannot be serialized");
		return false;
		}

	if ( ! info.cont )
		{
		StartSerialization();
		Write('i');
		}

	if ( info.incr )
		*info.incr = false;

	if ( ! id->Serialize(this, false, info) )
		{
		Error("failed");
		return false;
		}

	if ( (! info.incr) || (! *info.incr) )
		{
		WriteSeparator();

		if ( ! EndSerialization() )
			{
			Error(io->Error());
			return false;
			}
		}

	return true;
	}

bool Serializer::Serialize(const char* event, val_list* args)
	{
	assert(current_cache);

	SetErrorDescr(fmt("serializing event %s", event));

	StartSerialization();

	WriteOpenTag("event");
	int a = args->length();
	Write('e');
	Write(event, "name");
	Write(network_time, "time");
	Write(a);

	SerialInfo info;
	loop_over_list(*args, i) (*args)[i]->Serialize(this, true, info);

	WriteCloseTag("event");
	WriteSeparator();

	if ( ! EndSerialization() )
		{
		Error(io->Error());
		return false;
		}

	return true;
	}

int Serializer::Unserialize(bool block)
	{
	assert(current_cache);

	SetErrorDescr("unserializing");

	ChunkedIO::Chunk* chunk = 0;

	while ( ! chunk )
		{
		if ( ! io->Read(&chunk) )
			{
			if ( io->Eof() )
				return 0;
			Error(io->Error());
			return -1;
			}

		if ( ! chunk && ! block )
			return 0;
		}

	format->StartRead(chunk->data, chunk->len);

	char type;
	if ( ! format->Read(&type) )
		return false;

	bool result;
	switch ( type ) {
	case 'i':
		result = UnserializeID();
		break;

	case 'e':
		result = UnserializeEvent();
		break;

	default:
		Error(fmt("unknown serialization type %x", (int) type));
		result = false;
	}

	format->EndRead();

	delete [] chunk->data;
	delete chunk;

	return result;
	}

bool Serializer::UnserializeID()
	{
	SetErrorDescr("unserializing ID");

	ID* id = ID::Unserialize(this, false);

	if ( ! id )
		return false;

	GotID(id, id->ID_Val());

	if ( verbose )
		{
		ODesc d;
		id->Describe(&d);
		fprintf(stderr, "Got ID %s\n", d.Description());
		}

	return true;
	}

bool Serializer::UnserializeEvent()
	{
	const char* name;
	int len;
	double time;

	if ( ! (Read(&name, 0) && Read(&time) && Read(&len)) )
		return false;

	SetErrorDescr(fmt("unserializing event %s", name));

	FuncType* functype = 0;
	type_list* types = 0;

	EventHandler* handler = event_registry->Lookup(name);

	bool ignore = false;

	if ( handler )
		{
		functype = handler->FType();
		types = functype->ArgTypes()->Types();

		if ( types->length() != len )
			{
			Error("wrong number of arguments, ignoring");
			ignore = true;
			}
		}
	else
		{
		Error("unknown event, ignoring");
		ignore = true;
		}

	ODesc d;
	val_list* args = new val_list;
	for ( int i = 0; i < len; ++i )
		{
		Val* v = Val::Unserialize(this, 0, TYPE_ANY);

		if ( ! v )
			{
			delete [] name;
			return false;
			}

		if ( v->Type()->Tag() != (*types)[i]->Tag() && ! ignore )
			{
			Error("mismatch in argument types; ignoring");
			ignore = true;
			}

		if ( verbose && types && ! ignore )
			v->Describe(&d);

		args->append(v);
		}

	if ( ! ignore  )
		{
		if ( verbose )
			fprintf(stderr, "Event [%.8f] %s(%s)\n", time, name,
				types ? d.Description() : "<ignored>");

		GotEvent(name, time, handler, args);
		}
	else
		delete_vals(args);

	delete [] name;

	return true;
	}

void Serializer::Error(const char* str)
	{
	char buffer[1024];
	snprintf(buffer, sizeof(buffer), "%s%s%s",
		error_descr ? error_descr : "", error_descr ? ": " : "", str);
	ReportError(buffer);
	}

void Serializer::Warning(const char* str)
	{
	char buffer[1024];
	snprintf(buffer, sizeof(buffer), "%s%s%s\n",
		error_descr ? error_descr : "", error_descr ? ": " : "", str);
	fputs(buffer, stderr);
	}

SerializationCache::SerializationCache()
	{
	next_id = 1;
	object_map = new ObjectMap;
	}

SerializationCache::~SerializationCache()
	{
	Clear();
	delete object_map;
	}

SerializationCache::ObjectID SerializationCache::Register(const BroObj* obj, ObjectID id)
	{
	if ( id == NONE )
		id = next_id++;

	HashKey key(&obj, sizeof(obj));

	delete (ObjectID*) object_map->Remove(&key);

	object_map->Insert(&key, new ObjectID(id));

	if ( id >= id_table.size() )
		id_table.resize(id+1);

	id_table[id] = obj;
	Ref(const_cast<BroObj*>(obj)); // Not nice.

	return id;
	}

void SerializationCache::Clear()
	{
	ObjectID* id;
	IterCookie* cookie = object_map->InitForIteration();
	while ( (id = object_map->NextEntry(cookie)) )
		delete id;

	for ( unsigned int i = 0; i < id_table.size(); i++ )
		Unref(const_cast<BroObj*>(id_table[i])); // Not nice.

	next_id = 1;
	id_table.clear();

	delete object_map;
	object_map = new ObjectMap;
	}

FileSerializer::FileSerializer(SerializationFormat* format): Serializer(format)
	{
	file = 0;
	fd = -1;
	io = 0;
	SetCache(&cache);
	}

FileSerializer::~FileSerializer()
	{
	delete [] file;
	delete io;

	if ( fd >= 0 )
		close(fd);
	}

bool FileSerializer::Open(const char* file, bool pure)
	{
	if ( ! OpenFile(file, false) )
		return false;

	if ( pure )
		io->MakePure();

	if ( ! PrepareForWriting() )
		return false;

	return true;
	}

bool FileSerializer::Close()
	{
	CloseFile();
	return true;
	}

bool FileSerializer::OpenFile(const char* arg_file, bool readonly)
	{
	CloseFile();

	cache.Clear();

	file = copy_string(arg_file);
	fd = open(file, readonly ? O_RDONLY : O_WRONLY | O_CREAT | O_TRUNC, 0600);

	if ( fd < 0 )
		{
		if ( readonly && errno == ENOENT )
			{
			// Only an error if we want to print it.
			if ( IsPrintOnly() )
				{
				Error(fmt("%s does not exist", file));
				return false;
				}

			CloseFile();
			return true;
			}

		Error(fmt("can't open file %s for %s: %s",
				file, (readonly ? "reading" : "writing"),
				strerror(errno)));
		return false;
		}

	io = new ChunkedIOFd(fd);

	return io != 0;
	}

void FileSerializer::CloseFile()
	{
	if ( io )
		io->Flush();

	if ( fd >= 0 )
		close(fd);
	fd = -1;

	delete [] file;
	file = 0;

	delete io;
	io = 0;
	}

bool FileSerializer::PrepareForWriting()
	{
	if ( ! io->IsPure() )
		{
		// Write file header.
		uint32 magic = htonl(MAGIC);
		uint16 version = htons(DATA_FORMAT_VERSION);
		uint32 time = htonl((uint32)::time(0));

		if ( write(fd, &magic, sizeof(magic)) != sizeof(magic ) ||
		     write(fd, &version, sizeof(version)) != sizeof(version) ||
		     write(fd, &time, sizeof(time)) != sizeof(time))
			{
			Error(fmt("can't write file header to %s: %s",
					file, strerror(errno)));
			return false;
			}
		}

	return true;
	}

bool FileSerializer::ReadHeader()
	{
	uint32 magic;
	uint16 version;
	uint32 time;

	if ( read(fd, &magic, sizeof(magic)) != sizeof(magic ) ||
	     read(fd, &version, sizeof(version)) != sizeof(version) ||
	     read(fd, &time, sizeof(time)) != sizeof(time) )
		{
		Error(fmt("can't read file header from %s: %s",
				file, strerror(errno)));
		return false;
		}

	version = ntohs(version);
	time = ntohl(time);

	if ( verbose )
		{
		time_t teatime = (time_t) time;
		fprintf(stderr, "Date: %s", ctime(&teatime));
		}

	if ( magic != htonl(MAGIC) )
		{
		Error(fmt("%s is not a bro state file", file));
		CloseFile();
		return false;
		}

	if ( version != DATA_FORMAT_VERSION )
		{
		Error(fmt("wrong data format, expected version %d but got version %d", DATA_FORMAT_VERSION, version));
		CloseFile();
		return false;
		}

	return true;
	}

bool FileSerializer::Read(const char* file)
	{
	if ( ! OpenFile(file, true) )
		return false;

	fprintf( stderr, "Reading %s\n", file );

	if ( fd < 0 )
		// Not existent, but that's ok.
		return true;

	if ( ! ReadHeader() )
		return false;

	int i;
	while ( (i = Unserialize(true)) > 0 )
		;

	CloseFile();

	return i == 0;
	}

void FileSerializer::ReportError(const char* str)
	{
	run_time(str);
	}

void FileSerializer::GotID(ID* id, Val* val)
	{
	ODesc d;

	if ( id->AsType() )
		id->AsType()->Describe(&d);
	else if ( val )
		val->Describe(&d);
	else
		d.Add("<nil>");

	fprintf(stdout, "ID %s = %s\n", id->Name(), d.Description());
	}

void FileSerializer::GotEvent(const char* event, double time,
				EventHandlerPtr func, val_list* args)
	{
	ODesc d;
	for ( int i = 0; i < args->length(); i++ )
		(*args)[i]->Describe(&d);

	fprintf(stdout, "Event [%.8f] %s(%s)\n", time, event, d.Description());
	}

ConversionSerializer::ConversionSerializer(SerializationFormat* in,
						SerializationFormat* out)
: FileSerializer(in)
	{
	serout = new FileSerializer(out);
	}

ConversionSerializer::~ConversionSerializer()
	{
	delete serout;
	}

bool ConversionSerializer::Convert(const char* file_in, const char* file_out)
	{
	if ( ! serout->Open(file_out, true) )
		return false;

	if ( ! Read(file_in) )
		return false;

	if ( ! serout->Close() )
		return false;

	return true;
	}

void ConversionSerializer::GotEvent(const char* event, double time,
					EventHandlerPtr func, val_list* args)
	{
	serout->Serialize(event, args);
	}

void ConversionSerializer::GotID(ID* id, Val* val)
	{
	}

class IncrementalWriteTimer : public Timer {
public:
	IncrementalWriteTimer(double t) : Timer(t, TIMER_INCREMENTAL_WRITE)	{}
	virtual void Dispatch(double t, int is_expire)
		{ persistence_serializer->Proceed(!is_expire); }
};

PersistenceSerializer::PersistenceSerializer()
	{
	serial_running = false;
	dir = 0;
	current_id = -1;
	clear_list_when_finished = false;
	serial_ids = 0;
	serial_dir = 0;
	}

PersistenceSerializer::~PersistenceSerializer()
	{
	loop_over_list(persistent_ids, i)
		{
		Unref(persistent_ids[i]->id);
		delete [] persistent_ids[i]->file;
		delete persistent_ids[i];
		}

	delete [] dir;
	}

void PersistenceSerializer::RegisterID(ID* id)
	{
	if ( id->Type()->Tag() == TYPE_FUNC )
		{
		Error("can't register functions as persistent ID");
		return;
		}

	ID_Info* info = new ID_Info;
	info->id = id;
	Ref(info->id);

	info->file = copy_string(fmt("state.%s", id->Name()));
	info->body = 0;

	persistent_ids.append(info);
	}

bool PersistenceSerializer::CheckTimestamp(const char* file)
	{
	struct stat s;
	if ( stat(file, &s) < 0 )
		{
		Error(fmt("can't stat %s: %s", file, strerror(errno)));
		return false;
		}

	if ( ! S_ISREG(s.st_mode) )
		return false;

	bool changed = true;

	HashKey* key = new HashKey(file, strlen(file));
	time_t* t = files.Lookup(key);

	if ( t && *t >= s.st_mtime )
		changed = false;

	if ( ! t )
		files.Insert(key, new time_t(s.st_mtime));
	else
		*t = s.st_mtime;

	return changed;
	}

bool PersistenceSerializer::ReadAll()
	{
	assert(dir);

	struct stat s;
	if ( stat(dir, &s) < 0 )
		{
		if ( errno == ENOENT )
			return true;

		Error(fmt("can't stat %s: %s", dir, strerror(errno)));
		return false;
		}

	DIR* d = opendir(dir);
	if ( ! d )
		{
		Error(fmt("can't open directory %s: %s", dir, strerror(errno)));
		return false;
		}

	dirent* ent;
	while ( (ent = readdir(d)) )
		{
		// Ignore hidden files.
		if ( ent->d_name[0] == '.' )
			continue;

		const char* file = copy_string(fmt("%s/%s", dir, ent->d_name));

		if ( CheckTimestamp(file) )
			{
			Read(file);
			if ( unlink(file) < 0 )
				Error(fmt("can't delete %s: %s",
						file, strerror(errno)));
			}

		delete [] file;
		}

	closedir(d);

	return true;
	}

bool PersistenceSerializer::CheckDir(const char* dir)
	{
	struct stat s;
	if ( stat(dir, &s) < 0 )
		{
		if ( errno != ENOENT )
			{
			Error(fmt("can't stat %s: %s", dir, strerror(errno)));
			return false;
			}

		if ( mkdir(dir, 0700) < 0 )
			{
			Error(fmt("can't create directory %s: %s",
					dir, strerror(errno)));
			return false;
			}
		}

	else if ( ! S_ISDIR(s.st_mode) )
		{
		fprintf( stderr, dir );
		Error(fmt("%s exists but is not a directory", dir));
		return false;
		}

	return true;
	}

bool PersistenceSerializer::CheckSourceFile(const Location& l)
	{
	// Quite a simple heuristic, but should work ...
	// FIXME: There must be a better way than hard-coding this ...
	if ( strstr(l.filename, "/bro.init") ||
	     strstr(l.filename, "/pcap.bro") ||
	     strstr(l.filename, "/const.bif.bro") )
		return false;

	return true;
	}

bool PersistenceSerializer::WriteConfig(const char* dstdir)
	{
	if ( serial_running )
		return false;

	if ( dstdir )
		bro_logger->Log("Saving config...");

	if ( ! dstdir )
		dstdir = dir;

	if ( ! (CheckDir(dstdir) && CheckDir(fmt("%s/.tmp", dstdir))) )
		return false;

	// Build list of everything we need to store.
	assert(global_ids.length() == 0);
	PDict(ID)* globals = global_scope()->GetIDs();
	IterCookie* c = globals->InitForIteration();

	ID* id;
	while ( (id = globals->NextEntry(c)) )
		{
		if ( id->Type()->Tag() == TYPE_FUNC )
			{
			if ( ! id->HasVal() )
				continue;

			Func* f = id->ID_Val()->AsFunc();

			if ( f->GetKind() == Func::BUILTIN_FUNC )
				continue;	// can't serialize these

			// Split functions with multiple bodies (aka events)
			// into individual entries.
			if ( id->ID_Val()->AsFunc()->IsEvent() )
				{
				const vector<Stmt*>& bodies = f->GetBodies();
				for ( unsigned int j = 0; j < bodies.size(); ++j )
					{
					ID_Info* info = new ID_Info;

					const Location* l =
						bodies[j]->GetLocationInfo();

					if ( ! CheckSourceFile(*l) )
						continue;

					char* tmp = copy_string(l->filename);
					char* base = basename(tmp);
					info->file = copy_string(fmt("event.%s.%s.%d.bst", id->Name(), base, l->first_line));
					delete [] tmp;

					info->body = j;
					info->id = id;
					Ref(info->id);

					global_ids.append(info);
					}
				continue;
				}
			}

		if ( ! CheckSourceFile(*id->GetLocationInfo()) )
			continue;

		ID_Info* info = new ID_Info;
		info->id = id;
		Ref(info->id);

		const char* tag = "id";

		if ( id->AsType() )
			tag = "type";
		else if ( id->Type()->Tag() == TYPE_FUNC )
			tag = "func";

		info->file = copy_string(fmt("%s.%s.bst", tag, id->Name()));
		info->body = 0;	// 0 for functions that aren't events.

		global_ids.append(info);
		}

	no_content = dstdir == dir;
	serial_dir = dstdir;
	serial_ids = &global_ids;
	current_id = 0;
	id_suspended = false;
	clear_list_when_finished = true;

	return Proceed(false);
	}

bool PersistenceSerializer::WriteState(bool incremental)
	{
	assert(dir);

	if ( ! persistent_ids.length() )
		return true;

	if ( serial_running )
		return false; // Still a serialization running.

	if ( ! (CheckDir(dir) && CheckDir(fmt("%s/.tmp", dir))) )
		return false;

	bro_logger->Log("Checkpointing state...");

	no_content = false;
	serial_dir = dir;
	current_id = 0;
	serial_ids = &persistent_ids;
	id_suspended = false;
	clear_list_when_finished = false;

	return Proceed(incremental);
	}

bool PersistenceSerializer::Proceed(bool incremental)
	{
	// First, we serialize all data into the directory
	// <serial-dir>/.tmp/. Then we move the files one level
	// up. This makes sure that the old state keeps consistent
	// if something goes wrong.

	serial_running = true;

	bool success = false;

	while ( current_id < serial_ids->length() )
		{
		ID_Info* id = (*serial_ids)[current_id];

		if ( ! id_suspended )
			{
			if ( ! OpenFile(fmt("%s/.tmp/%s",
						serial_dir, id->file), false) )
				goto error;

			if ( ! PrepareForWriting() )
				goto error;

			++current_id;
			}

		if ( ! no_content )
			{
			SerialInfo info(incremental ?  &id_suspended : 0,
					id_suspended, id->body);

			if ( ! Serialize(id->id, info) )
				goto error;

			if ( info.incr &&
			     (current_id < serial_ids->length() ||
			      id_suspended) )
				{
				timer_mgr->Add(new IncrementalWriteTimer(network_time + state_write_delay));
				return true;
				}
			}
		}

	CloseFile();

	// Before we move the files to their new place, we first check if their
	// is any new data. If so, we read it in before overwriting it. When we do
	// the next checkpointing, the new data will be written.
	//
	// Yes, there's a race condition here.
	ReadAll();

	char oldname[PATH_MAX];
	char newname[PATH_MAX];

	char* oldp, *newp;

	oldp = oldname + snprintf(oldname, PATH_MAX, "%s/.tmp/", serial_dir );
	newp = newname + snprintf(newname, PATH_MAX, "%s/", serial_dir );

	success = true;

	loop_over_list(*serial_ids, i)
		{
		strncpy(oldp, (*serial_ids)[i]->file,
				PATH_MAX - (oldp - oldname) );
		strncpy(newp, (*serial_ids)[i]->file,
				PATH_MAX - (newp - newname) );

		oldname[PATH_MAX-1] = newname[PATH_MAX-1] = '\0';

		if ( rename(oldname, newname) < 0 )
			{
			Error(fmt("can't move %s to %s: %s",
					oldname, newname, strerror(errno)));
			success = false;
			}

		// Update the saved time-stamp.
		CheckTimestamp(newname);
		}

	if ( ! no_content )
		bro_logger->Log("Serialization done.");

error:
	serial_running = 0;

	if ( clear_list_when_finished )
		{
		loop_over_list(*serial_ids, i)
			{
			Unref((*serial_ids)[i]->id);
			delete [] (*serial_ids)[i]->file;
			delete (*serial_ids)[i];
			}

		serial_ids->clear();
		serial_ids = 0;
		}

	return success;
	}

void PersistenceSerializer::GotEvent(const char* event, double time,
					EventHandlerPtr func, val_list* args)
	{
	mgr.QueueEvent(func, args);
	}

void PersistenceSerializer::GotID(ID* id, Val* val)
	{
	// Install the ID.
	if ( ! id->IsGlobal() )
		{
		Error("persistent non-global id");
		delete id;
		return;
		}

	ID* current = global_scope()->Lookup(id->Name());

	if ( ! current )
		{
		global_scope()->Insert(id->Name(), id);
		if ( id->FindAttr(ATTR_PERSISTENT) )
			RegisterID(id);
		return;
		}

	if ( ! val )
		{
		Error("type already exists; ignored.");
		return;
		}

	// If it's an event handler for an already existing event, add
	// this to the current list of bodies.
	if ( val->Type()->Tag() == TYPE_FUNC &&
	     val->Type()->AsFuncType()->IsEvent() && current->HasVal() )
		{
		assert(val->AsFunc()->GetKind() == Func::BRO_FUNC);
		BroFunc* f = (BroFunc*) val->AsFunc();
		Func* cur = current->ID_Val()->AsFunc();
		const vector<Stmt*>& bodies = f->GetBodies();

		for( unsigned int i = 0; i < bodies.size(); i++ )
			cur->AddBody(bodies[i], 0, f->FrameSize());

		Unref(id);
		return;
		}

	// Change the value.
	// Type safety with respect to an existing global ID should
	// have already been assured.
	// FIXME: What about attributes? What about the other members?
	Ref(val);
	current->SetVal(val);

	if ( val && val->Type()->Tag() == TYPE_FUNC && val->AsFunc() )
		// We know that all our functions are in fact debug functions...
		((DebugFunc*) val->AsFunc())->SetID(current);

	Unref(id);
	}

EventPlayer::EventPlayer()
	{
	stream_time = 0;
	replay_time = 0;
	ne_time = 0;
	}

bool EventPlayer::Replay(const char* file)
	{
	if ( ! OpenFile(file, true) )
		return false;

	if ( fd < 0 )
		{
		Error(fmt("%s does not exist", file));
		return false;
		}

	return ReadHeader();
	}

EventPlayer::~EventPlayer()
	{
	CloseFile();
	}

void EventPlayer::GotEvent(const char* event, double time, EventHandlerPtr func, val_list* args)
	{
	ne_time = time;
	ne_handler = func;
	ne_args = args;
	}

double EventPlayer::NextTimestamp()
	{
	if ( ne_time )
		return ne_time;

	if ( ! io )
		return 0;

	// Read next event if we don't have one waiting.
	if ( ! ne_time )
		Unserialize();

	if ( ! ne_time )
		return 0;

	if ( ! network_time )
		{
		// Network time not initialized yet.
		stream_time = replay_time = ne_time;
		return ne_time;
		}

	if ( ! stream_time )
		{
		// Init base times.
		stream_time = ne_time;
		replay_time = network_time;
		}

	// Scale time.
	ne_time = ne_time - stream_time + network_time;
	return ne_time;
	}

void EventPlayer::DispatchNextEvent()
	{
	if ( ! ( io && ne_time ) )
		return;

	Event* event = new Event(ne_handler, ne_args);
	mgr.Dispatch(event);

	ne_time = 0;
	}

bool EventPlayer::IsActive()
	{
	if ( ! io )
		return false;

	return ! io->Eof();
	}
