// $Id: Serializer.h,v 1.3 2003/11/25 03:21:51 vern Exp $

#ifndef SERIALIZER_H
#define SERIALIZER_H

#include "ID.h"
#include "List.h"
#include "Expr.h"
#include "ChunkedIO.h"
#include "SerializationFormat.h"

class SerializationCache;
class SerialInfo;

class Serializer {
public:
	bool Serialize(ID* id, const SerialInfo& info);
	bool Serialize(const char* event, val_list* args);

	// Turns on output during unserialization.
	void Verbose()	{ verbose = 1; }

	// Tells the unserializers that all we want to do is printing the
	// state.  This may make some errors non-fatal.
	//
	// FIXME: The better approach would be to generally pass some
	// UnserialInfo to all Unserialize() functions (and, BTW, we
	// should make the use of SerialInfo more consistent, too.)

	void SetPrintOnly(bool arg_print_only)	{ print_only = arg_print_only; }
	bool IsPrintOnly() const	{ return print_only; }

	// Access to the current cache.
	SerializationCache* Cache()	{ return current_cache; }
	void SetCache(SerializationCache* cache)
		{ current_cache = cache; }

	// Input/output methods.

#define DECLARE_READ(type) \
	bool Read(type* v)	{ return format->Read(v); }

#define DECLARE_WRITE(type) \
	bool Write(type v, const char* tag = 0)	\
		{ return format->Write(v, tag); }

#define DECLARE_IO(type)	\
	DECLARE_READ(type)	\
	DECLARE_WRITE(type)

	DECLARE_IO(int)
	DECLARE_IO(uint16)
	DECLARE_IO(uint32)
	DECLARE_IO(char)
	DECLARE_IO(bool)
	DECLARE_IO(double)

	bool Read(const char** str, int* len)
		{ return format->Read(str, len); }

	bool Read(string* s);

	bool Write(const char* s, const char* tag = 0)
		{ return format->Write(s, tag); }
	bool Write(const char* buf, int len, const char* tag = 0)
		{ return format->Write(buf, len, tag); }
	bool Write(const string& s, const char* tag = 0)
		{ return format->Write(s.data(), s.size(), tag); }

	bool WriteOpenTag(const char* tag)
		{ return format->WriteOpenTag(tag); }
	bool WriteCloseTag(const char* tag)
		{ return format->WriteCloseTag(tag); }

	bool WriteSeparator()	{ return format->WriteSeparator(); }

	void Error(const char* msg);
	void Warning(const char* msg);

	void SetErrorDescr(const char* descr)
		{ delete [] error_descr; error_descr = copy_string(descr); }

protected:
	// Format defaults to binary serialization.
	Serializer(SerializationFormat* format = 0);
	virtual ~Serializer();

	// Reads next object.
	// If 'block' is true, wait until an object can be read.
	// Returns 0 if no more object available, -1 on error.
	int Unserialize(bool block = false);

	// Callback for error messages.
	virtual void ReportError(const char* msg) = 0;

	// Callback for unserialized IDs; id points to ID in global scope,
	// val is unserialized value.
	virtual void GotID(ID* id, Val* val) = 0;

	// Callback for unserialized events.
	virtual void GotEvent(const char* event, double time,
				EventHandlerPtr func, val_list* args) = 0;

	// Magic to recognize state files.
	static const uint32 MAGIC = 0x42525354;

	// This will be increased whenever there is an incompatible change
	// in the data format.
	static const int DATA_FORMAT_VERSION = 5;

	bool verbose;
	ChunkedIO* io;

private:
	void StartSerialization();
	bool EndSerialization();

	bool Serializer::UnserializeID();
	bool Serializer::UnserializeEvent();

	SerializationFormat* format;
	SerializationCache* current_cache;
	const char* error_descr;	// used in error messages

	bool print_only;
};

// We maintain a cache for some of the objects which have already been
// serialized. Each object gets an ID for later reference.
// Objects are identified by address. This is not perfect
// and could even be a little bit dangerous, so we have to make sure
// that we clear the cache for cases where addresses could change.
class SerializationCache {
public:
	typedef uint16 ObjectID;
	static const ObjectID NONE = 0;

	SerializationCache();
	~SerializationCache();

	ObjectID Register(const BroObj* obj, ObjectID id = NONE);

	const BroObj* Lookup(ObjectID id) const
		{ return id < id_table.size() ? id_table[id] : 0; }

	ObjectID Lookup(const BroObj* obj) const
		{
		HashKey key(&obj, sizeof(obj));
		ObjectID* id = object_map->Lookup(&key);
		return id ? *id : 0;
		}

	void Clear();

private:
	// Object cache.
	declare(PDict, ObjectID);
	typedef PDict(ObjectID) ObjectMap;
	typedef vector<const BroObj*> IDTable;

	IDTable id_table;
	ObjectMap* object_map;
	ObjectID next_id;
};

// Data to pass on to the serialization functions.
// (we have to declare it outside Serializer to make it
// forward-declarable... :-( ).
struct SerialInfo {
	SerialInfo(bool* arg_incr = 0, bool arg_cont = false, int arg_body = 0)
		{ incr = arg_incr; cont = arg_cont; body = arg_body; }

	// If non-nil, suspending the serialization is ok. Then
	// *incr has to be set to true to signal that the serialization
	// has not been finished yet.
	bool* incr;

	// If true, we're continuing a previously suspended serialization.
	bool cont;

	// If we serialize a function with multiple bodies (aka event handler),
	// this gives the number of the body we're refering to.
	int body;
};

// Write values/events to file or fd.
class FileSerializer : public Serializer {
public:
	FileSerializer(SerializationFormat* format = 0);
	virtual ~FileSerializer();

	// Opens the file for serialization.
	bool Open(const char* file, bool pure = false);
	bool Close();

	// Reads the file and prints content to stderr.
	bool Read(const char* file);

protected:
	virtual void ReportError(const char* msg);
	virtual void GotID(ID* id, Val* val);
	virtual void GotEvent(const char* event, double time,
				EventHandlerPtr func, val_list* args);

	bool OpenFile(const char* file, bool readonly);
	void CloseFile();
	bool ReadFile(const char* file);
	bool PrepareForWriting();
	bool ReadHeader();

	SerializationCache cache;
	const char* file;
	int fd;
};

// Converts from one serialization format into another.
class ConversionSerializer:public FileSerializer {
public:
	ConversionSerializer(SerializationFormat* in, SerializationFormat* out);
	virtual ~ConversionSerializer();

	bool Convert(const char* file_in, const char* file_out);

protected:
	virtual void GotID(ID* id, Val* val);
	virtual void GotEvent(const char* event, double time,
				EventHandlerPtr func, val_list* args);

	FileSerializer* serout;
};


// Implements persistance for values.
class PersistenceSerializer : public FileSerializer {
public:
	PersistenceSerializer();
	virtual ~PersistenceSerializer();

	// Define the directory where the data is to be stored.
	void SetDir(const char* arg_dir)	{ dir = copy_string(arg_dir); }

	// Registers the ID to be saved by WriteAll().
	void RegisterID(ID* id);

	// Reads all data that has been changed since last scan.
	bool ReadAll();

	// Writes all registered IDs. Existing data in
	// dir is overwritten. If incremental is true, may suspend
	// serialization after some amount of data is written.
	bool WriteState(bool incremental);

	// Writes Bro's configuration (w/o dynamic state).
	bool WriteConfig(const char* dstdir);

	// Returns true if last (incremental) WriteAll() hasn't finished yet.
	bool InProgress()	{ return serial_running; }

	// Performs (one step of incremental) serialization.
	bool Proceed(bool incremental);

protected:
	virtual void GotID(ID* id, Val* val);
	virtual void GotEvent(const char* event, double time,
				EventHandlerPtr func, val_list* args);

	// Make sure directory exists (creating it if not).
	bool CheckDir(const char* dir);

	// Checks if we want to store things from this location
	// (they may just clutter up our output).
	bool CheckSourceFile(const Location& l);

	// Returns true if it's a regular file and has a more recent timestamp
	// than last time we checked it.
	bool CheckTimestamp(const char* file);

	// Generates an error message, terminates current serialization,
	// and returns false.
	bool SerialError(const char* msg);

private:
	// Main state directory.
	const char* dir;

	struct ID_Info {
		ID* id;
		const char* file;	// File name w/o path.

		// For events: the body we want to reference.
		unsigned int body;
	};

	declare(PList, ID_Info);
	typedef PList(ID_Info) id_info_list;

	// Registered state IDs to store.
	id_info_list persistent_ids;

	// Temporary list of global IDs to store.
	id_info_list global_ids;

	// To keep track of files' modification times.
	declare(PDict, time_t);
	typedef PDict(time_t) file_map;
	file_map files;

	// State for incremental serialization.
	bool serial_running;
	bool no_content;
	bool id_suspended;
	bool clear_list_when_finished;
	int current_id;
	id_info_list* serial_ids;
	const char* serial_dir;
};


// Abstract interface class for external sources providing a stream of events.
class EventSource {
public:
	// Returns time of the oldest event (0 if none available).
	virtual double NextTimestamp() = 0;

	// Dispatches the oldest event and removes it.
	virtual void DispatchNextEvent() = 0;

	// Returns true if there are more events to expect from this source.
	virtual bool IsActive() = 0;
};

// Plays a file of events back (not implemeted yet).
class EventPlayer : public FileSerializer, public EventSource {
public:
	EventPlayer();
	virtual ~EventPlayer();

	bool Replay(const char* file);

	virtual double NextTimestamp();
	virtual void DispatchNextEvent();
	virtual bool IsActive();

protected:
	virtual void GotID(ID* id, Val* val)	{}
	virtual void GotEvent(const char* event, double time,
				EventHandlerPtr func, val_list* args);

	double stream_time;	// time of first captured event
	double replay_time;	// network time of replay start

	// Next event waiting to be dispatched.
	double ne_time;
	EventHandlerPtr ne_handler;
	val_list* ne_args;

};

extern EventPlayer* event_player;
extern FileSerializer* event_serializer;
extern PersistenceSerializer* persistence_serializer;

declare(PList, EventSource);
typedef PList(EventSource) event_source_list;
extern event_source_list external_event_sources;

// Some helpers for pointers which may be nil.
#define SERIALIZE_OPTIONAL(s, ptr) \
	{	\
	if ( ptr )	\
		{	\
		if ( ! ( s->Write(true) && ptr->Serialize(s) ) )	\
			return false;	\
		}	\
	else	\
		if ( ! s->Write(false) )	\
			return false;	\
	}

#define SERIALIZE_OPTIONAL_VAL(s, ptr, t) \
	{	\
	if ( ptr )	\
		{	\
		SerialInfo no_incremental;	\
		if ( ! ( s->Write(true) && ptr->Serialize(s, t, no_incremental) ) )	\
			return false;	\
		}	\
	else	\
		if ( ! s->Write(false) )	\
			return false;	\
	}

#define UNSERIALIZE_OPTIONAL(s, dst, unserialize)	\
	{	\
	bool has_it;	\
	if ( ! s->Read(&has_it) )	\
		return false;	\
	if ( has_it )	\
		{	\
		dst = unserialize;	\
		if ( ! dst )	\
			return false;	\
		}	\
	else	\
		dst = 0;	\
	}

#define UNSERIALIZE_OPTIONAL_STATIC(s, dst, unserialize, del)	\
	{	\
	bool has_it;	\
	if ( ! s->Read(&has_it) )	\
			{	\
			delete del;	\
			return 0;	\
			}	\
	if ( has_it )	\
		{	\
		dst = unserialize;	\
		if ( ! dst )	\
			{	\
			delete del;	\
			return 0;	\
			}	\
		}	\
	else	\
		dst = 0;	\
	}

#endif
