// $Id: Stats.cc,v 1.2 2003/11/14 08:14:20 vern Exp $

#include "Conn.h"
#include "File.h"
#include "Event.h"
#include "NetVar.h"
#include "Sessions.h"
#include "Stats.h"
#include "Scope.h"

#ifdef HAVE_MALLINFO
# include <malloc.h>
#else
# include <sys/resource.h>
#endif

class StatsTimer : public Timer {
public:
	StatsTimer(double t, StatsLogger* l, double i) : Timer(t, TIMER_STATS)
		{
		logger = l;
		interval = i;
		}

	virtual void Dispatch(double t, int is_expire)
		{
		if ( active )
			logger->Log();

		// Reinstall timer.
		if ( ! is_expire )
			timer_mgr->Add(new StatsTimer(network_time + interval,
							logger, interval));
		}

protected:
	double interval;
	StatsLogger* logger;
};

StatsLogger::StatsLogger(BroFile* arg_file, double interval)
	{
	file = arg_file;
	timer_mgr->Add(new StatsTimer(1, this, interval));
	}

StatsLogger::~StatsLogger()
	{
	file->Close();
	}

void StatsLogger::Log()
	{
	file->Write(fmt("%.8f ------------------------\n", network_time));

	// Memory information.
#ifdef HAVE_MALLINFO
	const struct mallinfo& mi = mallinfo();
	unsigned int malloced = mi.uordblks;
	unsigned int total = mi.arena;
#else
	// This gives bogus results on Linux.
	struct rusage r;
	getrusage(RUSAGE_SELF, &r);
	unsigned int malloced = 0;
	unsigned int total = r.ru_maxrss * 1024; // at least on FreeBSD it's in KB
#endif

	static unsigned int first_total = 0;

	file->Write(fmt("%.8f Mem: total=%dK total_adj=%dK malloced: %dK\n",
		network_time, total / 1024, (total - first_total) / 1024,
		malloced / 1024));

	if ( first_total == 0 )
		first_total = total;

	// Connection information.
	file->Write(fmt("%.8f Conns: total=%d current=%d/%d mem=%dK avg=%.1f\n",
		network_time,
		Connection::TotalConnections(),
		Connection::CurrentConnections(),
		sessions->CurrentConnections(),
		sessions->ConnectionMemoryUsage() / 1024,
		sessions->ConnectionMemoryUsage() /
			double(sessions->CurrentConnections())
		));

	SessionStats s;
	sessions->GetStats(s);

	file->Write(fmt("%.8f Conns: tcp=%d/%d udp=%d/%d icmp=%d/%d\n",
		network_time,
		s.num_TCP_conns, s.max_TCP_conns,
		s.num_UDP_conns, s.max_UDP_conns,
		s.num_ICMP_conns, s.max_ICMP_conns
		));

	// Reassembler.
	file->Write(fmt("%.8f Total reassembler data: %dK\n", network_time,
		Reassembler::TotalMemoryAllocation() / 1024));

	// Timer information (the used memory is a wild guess).
	file->Write(fmt("%.8f Timers: current=%d max=%d mem=%dK\n",
		network_time,
		timer_mgr->Size(), timer_mgr->PeakSize(),
		timer_mgr->Size() * (sizeof(Timer*) + sizeof(Timer)) / 1024
	));

	unsigned int* current_timers = TimerMgr::CurrentTimers();
	for( int i = 0; i < NUM_TIMER_TYPES; i++ )
		{
		if ( current_timers[i] )
			file->Write(fmt("%.8f         %s = %d\n", network_time,
					timer_type_to_string((TimerType) i),
					current_timers[i]));
			}

	// Script level state.
	unsigned int size, mem = 0;
	PDict(ID)* globals = global_scope()->Vars();
	IterCookie* c = globals->InitForIteration();

	file->Write(fmt("%.8f Global_sizes > 10k: %dK\n",
			network_time, mem / 1024));

	ID* id;
	while ( (id = globals->NextEntry(c)) )
		if ( id->HasVal() )
			{
			Val* v = id->ID_Val();

			size = id->ID_Val()->MemoryAllocation();
			mem += size;

			if ( size > 10 * 1024 )
				{
				file->Write(fmt("%.8f                %s = %dK",
					network_time, id->Name(), size / 1024));

				if ( v->Type()->Tag() == TYPE_TABLE )
					file->Write(fmt(" (%d entries)\n",
						v->AsTable()->Length()));
				else
					file->Write("\n");
				}
			}

	file->Write(fmt("%.8f Global_sizes total: %dK\n",
			network_time, mem / 1024));

	// Create an event so that scripts can log their information, too.
	// (and for consistency we dispatch it *now*)
	if ( statistics_update )
		{
		val_list* vl = new val_list;
		vl->append(new Val(file));
		mgr.Dispatch(new Event(statistics_update, vl));
		}
	}
